#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Rewrite and/or check time axis of MIP NetCDF files.

"""

import logging
import os
import re
import sys
from datetime import datetime
from functools import wraps
from multiprocessing import Pool
from textwrap import fill

import numpy as np

import db
from constants import *
from handler import File
from nctime.utils import time, utils


class ProcessingContext(object):
    """
    Encapsulates the following processing context/information for main process:

    +-----------------------+-------------+---------------------------------+
    | Attribute             | Type        | Description                     |
    +=======================+=============+=================================+
    | *self*.directory      | *str*       | Variable to scan                |
    +-----------------------+-------------+---------------------------------+
    | *self*.write          | *boolean*   | True if write mode              |
    +-----------------------+-------------+---------------------------------+
    | *self*.force          | *boolean*   | True if force writing           |
    +-----------------------+-------------+---------------------------------+
    | *self*.verbose        | *boolean*   | True if verbose mode            |
    +-----------------------+-------------+---------------------------------+
    | *self*.project        | *str*       | MIP project                     |
    +-----------------------+-------------+---------------------------------+
    | *self*.db             | *str*       | Database path                   |
    +-----------------------+-------------+---------------------------------+
    | *self*.threads        | *int*       | Maximal threads number          |
    +-----------------------+-------------+---------------------------------+
    | *self*.checksum_type  | *str*       | The checksum type               |
    +-----------------------+-------------+---------------------------------+
    | *self*.pattern        | *re object* | Filename regex pattern          |
    +-----------------------+-------------+---------------------------------+
    | *self*.tunits_default | *str*       | Default time units              |
    +-----------------------+-------------+---------------------------------+
    | *self*.ref            | *str*       | First filename as reference     |
    +-----------------------+-------------+---------------------------------+
    | *self*.variable       | *str*       | MIP variable                    |
    +-----------------------+-------------+---------------------------------+

    :param *ArgumentParser* args: Parsed command-line arguments
    :returns: The processing context
    :rtype: *ProcessingContext*

    """

    def __init__(self, args):
        self.directory = args.directory
        self.write = args.write
        self.force = args.force
        self.verbose = args.v
        self.project = args.project
        self.db = utils.path_switcher('db', args, default='{0}/{1}'.format(os.getcwd(), 'timeaxis.db'))
        cfg = utils.config_parse(args.i, self.project)
        self.threads = args.max_threads
        if self.threads == 0:
            if cfg.has_option('DEFAULT', 'max_threads') and cfg.get('DEFAULT', 'max_threads') != '':
                self.threads = cfg.get('DEFAULT', 'max_threads')
            else:
                self.threads = MAX_THREADS_DEFAULT
        self.checksum_type = str(cfg.get('DEFAULT', 'checksum_type'))
        self.pattern = utils.translate_filename_format(cfg, self.project)
        if cfg.has_option('time_units_default', self.project):
            self.tunits_default = cfg.get(self.project, 'time_units_default')
        else:
            self.tunits_default = None
        self.ref = self.get_files_list().next()
        self.variable = unicode(self.ref.split('_')[0])

    def get_files_list(self):
        """
        Yields all files into a directory.

        :returns: A file iterator
        :rtype: *iter*

        """
        for filename in sorted(os.listdir(self.directory)):
            if not os.path.isfile(os.path.join(self.directory, filename)):
                logging.warning('{0} is not a file and was skipped'.format(filename))
                continue
            if any(test_fx in filename for test_fx in ['_fx_', '_fixed_', '_fx.', '_fixed.']):
                logging.warning('STOP because "fixed" frequency has no time axis')
                sys.exit(0)
            if not re.match(self.pattern, filename):
                logging.warning('{0} has invalid filename and was skipped'.format(filename))
                continue
            yield filename


def counted(fct):
    """
    Decorator used to count all file process calls.

    :param callable fct: The function to monitor
    :returns: A wrapped function with a ``.called`` attribute
    :rtype: *callable*

    """

    @wraps(fct)  # Convenience decorator to keep the file_process docstring
    def wrapper(*args, **kwargs):
        wrapper.called += 1
        return fct(*args, **kwargs)

    wrapper.called = 0
    wrapper.__name__ = fct.__name__
    return wrapper


@counted
def process(inputs):
    """
    time_axis_processing(inputs)

    Time axis process that\:
     * Deduces start and end dates from filename,
     * Rebuilds the theoretical time axis (using frequency, calendar, etc.),
     * Compares the theoretical time axis with the time axis from the file,
     * Compares the last theoretical date with the end date from the filename,
     * Checks if the expected time units keep unchanged,
     * Checks the squareness and the consistency of time boundaries,
     * Rewrites (with ``--write`` mode) the new time axis,
     * Computes the new checksum if modified,
     * Traceback the status.

    :param tuple inputs: A tuple with the filename, the processing context and the time initialization
    :returns: The updated file handler instance
    :rtype: *nctime.axis.handler.File*

    """
    # Extract processing context and time initialization from input tuple
    filename, ctx, init = inputs

    # Instantiate file handle from input tuple
    handler = File(directory=ctx.directory,
                   filename=filename,
                   has_bounds=init.has_bounds)

    # Extract start and end dates from filename
    start, _, _ = handler.get_start_end_dates(pattern=ctx.pattern,
                                              frequency=init.frequency,
                                              units=init.funits,
                                              calendar=init.calendar)

    # Rebuild a theoretical time axis with high precision
    handler.time_axis_rebuilt = handler.build_time_axis(start=start,
                                                        inc=time.time_inc(init.frequency)[0],
                                                        input_units=init.funits,
                                                        output_units=init.tunits,
                                                        calendar=init.calendar,
                                                        is_instant=init.is_instant)

    # Check consistency between last time date and end date from filename
    if handler.last_date != handler.end_date:
        # Rebuild a theoretical time axis with low precision
        handler.time_axis_rebuilt = handler.build_time_axis(start=utils.trunc(start, 5),
                                                            inc=time.time_inc(init.frequency)[0],
                                                            input_units=init.funits,
                                                            output_units=init.tunits,
                                                            calendar=init.calendar,
                                                            is_instant=init.is_instant)
        if handler.last_date != handler.end_date:
            handler.status.append('003')
            logging.error('{0} - 003 - Last time step differs from end date from filename'.format(filename))
            if ctx.verbose:
                axe = fill(' | '.join(map(str, handler.time_axis.tolist())), width=100)
                logging.info('{0} - Time axis:\n {1}'.format(filename, axe))
                axe = fill(' | '.join(map(str, handler.time_axis_rebuilt.tolist())), width=100)
                logging.info('{0} - Theoretical axis:\n {1}'.format(filename, axe))
            return handler

    # Check consistency between instant time and time boundaries
    if init.is_instant and init.has_bounds:
        handler.status.append('004')
        logging.error('{0} - 004 - An instantaneous time axis should not embed time boundaries'.format(filename))

    # Check consistency between averaged time and time boundaries
    if not init.is_instant and not init.has_bounds:
        handler.status.append('005')
        logging.error('{0} - 005 - An averaged time axis should embed time boundaries'.format(filename))

    # Check time axis squareness
    if not np.array_equal(handler.time_axis_rebuilt, handler.time_axis):
        handler.status.append('001')
        logging.error('{0} - 001 - Mistaken time axis over one or several time steps'.format(filename))
    else:
        handler.status.append('000')
        logging.info('{0} - Time axis seems OK'.format(filename))

    # Check time boundaries squareness if needed
    if init.has_bounds:
        handler.time_bounds_rebuilt = handler.build_time_bounds(start=utils.trunc(start, 5),
                                                                inc=time.time_inc(init.frequency)[0],
                                                                input_units=init.funits,
                                                                output_units=init.tunits,
                                                                calendar=init.calendar)
        if not np.array_equal(handler.time_bounds_rebuilt, handler.time_bounds):
            handler.status.append('006')
            logging.error('{0} - 006 - Mistaken time bounds over one or several time steps'.format(filename))

    # Check consistency between time units
    if init.tunits != handler.time_units:
        handler.status.append('002')
        logging.error('{0} - 002 - Time units must be unchanged for the same dataset.'.format(filename))

     # Check consistency between time units
    if init.calendar != handler.calendar:
        handler.status.append('007')
        logging.error('{0} - 007 - Calendar must be unchanged for the same dataset.'.format(filename))

    # Rewrite time axis depending on checking
    if (ctx.write or ctx.force) and set(['004']).intersection(set(handler.status)):
        # Delete time bounds and bounds attribute from file if write of force mode
        handler.nc_var_delete(variable='time_bnds')
        handler.nc_att_delete(attribute='bounds', variable='time')

    # Rewrite time axis depending on checking
    if (ctx.write and set(['001', '002', '006', '007']).intersection(set(handler.status))) or ctx.force:
        handler.nc_var_overwrite('time', handler.time_axis_rebuilt)
        handler.nc_att_overwrite('units', 'time', init.tunits)
        handler.nc_att_overwrite('calendar', 'time', init.calendar)
        # Rewrite time boundaries if needed
        if init.has_bounds:
            handler.nc_var_overwrite('time_bnds', handler.time_bounds_rebuilt)

    # Compute checksum at the end of all modifications and after closing file
    if (ctx.write or ctx.force) and set(['001', '002', '004', '006', '007']).intersection(set(handler.status)):
        handler.new_checksum = handler.checksum(ctx.checksum_type)

    # Return file status
    return handler


def yield_inputs(ctx, tinit):
    """
    Yields all files to process within tuples with the processing context and the time initialization.

    :param nctime.axis.main.ProcessingContext ctx: The processing context
    :param nctime.utils.time.TimeInit tinit: The time initialization context
    :returns: Attach the processing context and the time initialization to a file processing as an iterator of tuples
    :rtype: *iter*

    """
    for filename in ctx.get_files_list():
        yield filename, ctx, tinit


def wrapper(inputs):
    """
    Transparent wrapper for pool map.

    :param tuple inputs: A tuple with the file path and the processing context
    :returns: The :func:`process` call
    :rtype: *callable*
    :raises Error: When a thread-process failed preserving its traceback

    """
    filename, ctx, _ = inputs

    try:
        return process(inputs)
    except Exception as e:
        # Use verbosity to raise the whole threads traceback errors
        if not ctx.verbose:
            logging.error('{0} skipped\n{1}: {2}'.format(filename, e.__class__.__name__, e.message))
        else:
            logging.exception('{0} failed'.format(filename))
        return None


def main(args):
    """
    Main process that\:
     * Instanciates processing context,
     * Defines the referenced time properties,
     * Instanciates threads pools,
     * Prints or logs the time axis diagnostics.

    :param ArgumentParser args: Parsed command-line arguments

    """
    # Instantiate processing context from command-line arguments
    ctx = ProcessingContext(args)
    logging.info('Time axis diagnostic started for {0}'.format(ctx.directory))
    # Set driving time properties (e.g., calendar, frequency and time units) from first file in directory
    tinit = time.TimeInit(ctx)
    # Process
    pool = Pool(int(ctx.threads))
    handlers = pool.imap(wrapper, yield_inputs(ctx, tinit))
    status = [] ; counter = 0
    # Persist diagnostics into database
    if ctx.db:
        # Check if database exists
        if not os.path.isfile(ctx.db):
            logging.warning('Database does not exist')
            db.create(ctx.db)
    # Commit each diagnostic as a new entry
    for handler in handlers:
        status.extend(handler.status)
        counter += 1
        if ctx.db:
            diagnostic = dict()
            diagnostic['creation_date'] = datetime.now()
            diagnostic.update(ctx.__dict__)
            diagnostic.update(tinit.__dict__)
            diagnostic.update(handler.__dict__)
            diagnostic['status'] = ','.join(handler.status)
            db.insert(ctx.db, diagnostic)
            logging.info('{0} - Diagnostic persisted into database'.format(handler.filename))
        if ctx.verbose:
            logging.info('-> Filename: {0}'.format(handler.filename))
            logging.info('Start: {0}'.format(handler.start_date))
            logging.info('End: {0}'.format(handler.end_date))
            logging.info('Last: {0}'.format(handler.last_date))
            logging.info('Time steps: {0}'.format(handler.length))
            logging.info('Is instant: {0}'.format(tinit.is_instant))
            logging.info('-> Time axis:')
            logging.info('{0}'.format(fill(' | '.join(map(str, handler.time_axis.tolist())), width=100)))
            logging.info('-> Theoretical axis:')
            logging.info('{0}'.format(fill(' | '.join(map(str, handler.time_axis_rebuilt.tolist())), width=100)))
    # Close tread pool
    pool.close()
    pool.join()
    if any([s in EXIT_ERRORS for s in status]):
        logging.error('Some time axis should be corrected manually ({0} files scanned)'.format(counter))
        sys.exit(1)
    else:    
        logging.info('Time diagnostic completed ({0} files scanned)'.format(counter))
        sys.exit(0)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Constants used in this module.

"""

# Database connexion timeout (in sec)
TIMEOUT = 120

# Default maximum threads number
MAX_THREADS_DEFAULT = 1

# Errors that return exit code = 1
EXIT_ERRORS = ['003', '005']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
   :platform: Unix
   :synopsis: Highlight chunked NetCDF files producing overlap in a time series.

"""

import logging
import os
import re
import sys

import networkx as nx

from nctime.utils import time, utils


class ProcessingContext(object):
    """
    Encapsulates the following processing context/information for main process:

    +-----------------------+-------------+---------------------------------+
    | Attribute             | Type        | Description                     |
    +=======================+=============+=================================+
    | *self*.directory      | *str*       | Variable to scan                |
    +-----------------------+-------------+---------------------------------+
    | *self*.mip            | *str*       | The MIP table from DRS          |
    +-----------------------+-------------+---------------------------------+
    | *self*.remove         | *boolean*   | True if remove mode             |
    +-----------------------+-------------+---------------------------------+
    | *self*.verbose        | *boolean*   | True if verbose mode            |
    +-----------------------+-------------+---------------------------------+
    | *self*.subtree        | *boolean*   | True for sub-period use         |
    +-----------------------+-------------+---------------------------------+
    | *self*.project        | *str*       | MIP project                     |
    +-----------------------+-------------+---------------------------------+
    | *self*.pattern        | *re object* | Filename regex pattern          |
    +-----------------------+-------------+---------------------------------+
    | *self*.tunits_default | *str*       | Default time units              |
    +-----------------------+-------------+---------------------------------+
    | *self*.ref            | *str*       | First filename as reference     |
    +-----------------------+-------------+---------------------------------+
    | *self*.variable       | *str*       | MIP variable                    |
    +-----------------------+-------------+---------------------------------+

    :param *ArgumentParser* args: Parsed command-line arguments
    :returns: The processing context
    :rtype: *ProcessingContext*

    """

    def __init__(self, args):
        self.directory = args.directory
        self.mip = args.mip
        self.remove = args.remove
        self.verbose = args.v
        self.subtree = args.subtree
        self.project = args.project
        cfg = utils.config_parse(args.i, self.project)
        self.pattern = utils.translate_filename_format(cfg, self.project)
        if cfg.has_option('time_units_default', self.project):
            self.tunits_default = cfg.get(self.project, 'time_units_default')
        else:
            self.tunits_default = None
        self.ref = self.get_files_list().next()
        self.variable = unicode(self.ref.split('_')[0])

    def get_files_list(self):
        """
        Yields all files into a directory.

        :returns: A file iterator
        :rtype: *iter*

        """
        for filename in sorted(os.listdir(self.directory)):
            if not os.path.isfile(os.path.join(self.directory, filename)):
                logging.warning('{0} is not a file and was skipped'.format(filename))
                continue
            if any(test_fx in filename for test_fx in ['_fx_', '_fixed_', '_fx.', '_fixed.']):
                logging.warning('STOP because "fixed" frequency has no time axis')
                sys.exit(0)
            if not re.match(self.pattern, filename):
                logging.warning('{0} has invalid filename and was skipped'.format(filename))
                continue
            yield filename


def get_overlaps(nodes, shortest, overlaps=True):
    """
    Yields all matching or overlapping (default) files.

    :param list nodes: The directed graph nodes as a list of tuples
    :param list shortest: The most consecutive files list (from `nx.DiGraph()`)
    :param boolean overlaps: True if overlaps are returned
    :returns: The filenames
    :rtype: *list*

    """
    match = []
    for i in range(len(shortest) - 1):
        match.append(zip(*nodes)[0][zip(*nodes)[1].index(shortest[i]) and
                                    zip(*nodes)[3].index(shortest[i + 1])])
    # Find overlapping filenames
    if overlaps:
        return list(set(zip(*nodes)[0]).symmetric_difference(set(match)))
    else:
        return match


def main(args):
    """
    Main process that\:
     * Instanciates processing context,
     * Deduces start, end and next date from each filenames,
     * Builds the DiGraph,
     * Detects the shortest path between dates if exists,
     * Removes the overlapping files.

    :param ArgumentParser args: Parsed command-line arguments

    """
    # Instantiate processing context from command-line arguments or SYNDA job dictionnary
    ctx = ProcessingContext(args)
    logging.info('Overlap diagnostic started for {0}'.format(ctx.directory))
    # Set driving time properties (e.g., calendar, frequency and time units) from first file in directory
    tinit = time.TimeInit(ctx)
    # DiGraph creation
    graph = nx.DiGraph()
    graph.clear()
    nodes = []
    if ctx.verbose:
        logging.info('{0} | {1} | {2} | {3}'.format('File'.center(len(ctx.ref) + 2),
                                                    'Start'.center(19),
                                                    'End'.center(19),
                                                    'Next'.center(19)))
    for filename in ctx.get_files_list():
        start, end, nxt = time.get_start_end_dates_from_filename(filename=filename,
                                                                 pattern=ctx.pattern,
                                                                 frequency=tinit.frequency,
                                                                 calendar=tinit.calendar)
        if ctx.verbose:
            logging.info('{0} | {1} | {2} | {3}'.format(filename.center(len(filename) + 2),
                                                        time.date2str(start).center(19),
                                                        time.date2str(end).center(19),
                                                        time.date2str(nxt).center(19)))
        # A node is defined by its start, end and next dates as a tuple.
        start, end, nxt = time.dates2str([start, end, nxt], sep=False)
        nodes.append((filename, int(start), int(end), int(nxt)))
        # Add nodes and edges
        graph.add_edge(int(start), int(nxt))
    # Walk through the graph
    try:
        # Find shortest path between oldest and latest dates
        shortest = nx.shortest_path(graph,
                                    source=min(zip(*nodes)[1]),
                                    target=max(zip(*nodes)[3]))
        logging.info('Shortest path found')
        overlaps = get_overlaps(nodes, shortest, overlaps=True)
    except nx.NetworkXNoPath, e:
        if ctx.subtree:
            # Find shortest path of the subtree instead
            pred, dist = nx.bellman_ford(graph, min(zip(*nodes)[1]))
            latest = utils.find_key(dist, max(dist.values()))
            shortest = nx.shortest_path(graph,
                                        source=min(zip(*nodes)[1]),
                                        target=latest)
            logging.warning('Shortest path found on the longest subtree from start:')
            for filename in get_overlaps(nodes, shortest, overlaps=False):
                logging.warning(filename)
            overlaps = get_overlaps(nodes, shortest, overlaps=True)
        else:
            logging.warning('No shortest path found: {0}'.format(str(e)))
            overlaps = None
    # Print results
    if not overlaps:
        logging.info('No overlapping files')
    else:
        logging.warning('Overlapping files:')
        for filename in overlaps:
            logging.warning(filename)
    if ctx.remove:
        for filename in overlaps:
            os.remove('{0}/{1}'.format(ctx.directory, filename))
        logging.warning('{0} overlapping files removed'.format(len(overlaps)))
    logging.info('Overlap diagnostic completed')



#!/usr/bin/env python
"""
.. module:: nctime.axis.handler.py
   :platform: Unix
   :synopsis: File handler for time axis.

"""

import os
import gc
from uuid import uuid4

import nco
import netCDF4
import numpy as np

from exceptions import *
from nctime.utils import time
from nctime.utils.exceptions import *


class File(object):
    """
    Handler providing methods to deal with file processing.
    It encapsulates the following file diagnostic to print or save:

    +------------------+-----------+---------------------------------------------------+
    | Attribute        | Type      | Description                                       |
    +==================+===========+===================================================+
    | *self*.directory | *str*     | Variable to scan                                  |
    +------------------+-----------+---------------------------------------------------+
    | *self*.file      | *str*     | Filename                                          |
    +------------------+-----------+---------------------------------------------------+
    | *self*.start     | *str*     | Start date from filename                          |
    +------------------+-----------+---------------------------------------------------+
    | *self*.end       | *str*     | End date from filename                            |
    +------------------+-----------+---------------------------------------------------+
    | *self*.last      | *str*     | Last date from theoretical time axis              |
    +------------------+-----------+---------------------------------------------------+
    | *self*.steps     | *int*     | Time axis length                                  |
    +------------------+-----------+---------------------------------------------------+
    | *self*.instant   | *boolean* | True if instantaneous time axis                   |
    +------------------+-----------+---------------------------------------------------+
    | *self*.calendar  | *str*     | NetCDF calendar attribute                         |
    +------------------+-----------+---------------------------------------------------+
    | *self*.units     | *str*     | Expected time units                               |
    +------------------+-----------+---------------------------------------------------+
    | *self*.control   | *list*    | Errors status                                     |
    +------------------+-----------+---------------------------------------------------+
    | *self*.bnds      | *boolean* | True if time bounds excpected and not mistaken    |
    +------------------+-----------+---------------------------------------------------+
    | *self*.checksum  | *str*     | New checksum if modified file                     |
    +------------------+-----------+---------------------------------------------------+
    | *self*.axis      | *array*   | Theoretical time axis                             |
    +------------------+-----------+---------------------------------------------------+
    | *self*.time      | *array*   | Time axis from file                               |
    +------------------+-----------+---------------------------------------------------+

    :returns: The axis status
    :rtype: *File*

    """

    def __init__(self, directory, filename, has_bounds):
        self.directory, self.filename = directory, filename
        # Retrieve file full path
        self.ffp = os.path.join(self.directory, self.filename)
        # Start/end period dates from filename
        self.start_date = None
        self.end_date = None
        self.last_date = None
        # Get time axis length
        try:
            f = netCDF4.Dataset(self.ffp, 'r')
        except IOError:
            raise InvalidNetCDFFile(self.ffp)
        self.length = f.variables['time'].shape[0]
        # Get time boundaries
        if has_bounds:
            self.time_bounds = f.variables['time_bnds'][:, :]
        else:
            self.time_bounds = None
        # Get time axis
        self.time_axis = f.variables['time'][:]
        # Get time units from file
        self.time_units = f.variables['time'].units
        # Get calendar from file
        self.calendar = f.variables['time'].calendar        
        f.close() ; del f ; gc.collect()
        self.time_axis_rebuilt = None
        self.time_bounds_rebuilt = None
        self.status = list()
        self.new_checksum = None

    def get_start_end_dates(self, pattern, frequency, units, calendar):
        """
        Wraps and records :func:`get_start_end_dates_from_filename` results.

        :param re Object pattern: The filename pattern as a regex (from `re library \
        <https://docs.python.org/2/library/re.html>`_).
        :param str frequency: The time frequency
        :param str units: The proper time units
        :param str calendar: The NetCDF calendar attribute
        :returns: Start and end dates as number of days since the referenced date
        :rtype: *float*

        """
        dates = time.get_start_end_dates_from_filename(filename=self.filename,
                                                       pattern=pattern,
                                                       frequency=frequency,
                                                       calendar=calendar)
        self.start_date, self.end_date, _ = time.dates2str(dates)
        return time.date2num(dates, units=units, calendar=calendar)

    def checksum(self, checksum_type):
        """
        Does the checksum by the Shell avoiding Python memory limits.

        :param str checksum_type: Checksum type
        :returns: The checksum
        :rtype: *str*
        :raises Error: If the checksum fails

        """
        checksum_client = {'SHA256': 'sha256sum',
                           'MD5':    'md5sum'}
        assert (checksum_type in checksum_client.keys()), 'Invalid checksum type'
        try:
            shell = os.popen("{0} {1} | awk -F ' ' '{{ print $1 }}'".format(checksum_client[checksum_type],
                                                                            self.ffp),
                             'r')
            return shell.readline()[:-1]
        except:
            raise ChecksumFail(checksum_type, self.ffp)

    def nc_var_delete(self, variable):
        """
        Delete a NetCDF variable using NCO operators.
        A unique filename is generated to avoid multithreading errors.
        To overwrite the input file, the source file is dump using the ``cat`` Shell command-line
        to avoid Python memory limit.

        :param str variable: The variable to delete
        :raises Error: If the deletion failed

        """
        # Generate unique filename
        fftmp = '{0}/{1}{2}'.format(self.directory, str(uuid4()), '.nc')
        try:
            nc = nco.Nco()
            nc.ncks(input=self.ffp,
                    output=fftmp,
                    options='-x -O -v {0}'.format(variable))
            os.popen("cat {0} > {1}".format(fftmp, self.ffp), 'r')
            os.remove(fftmp)
        except:
            os.remove(fftmp)
            raise NetCDFVariableRemoveFail(variable, self.ffp)

    def nc_att_delete(self, variable, attribute):
        """
        Delete a NetCDF dimension attribute using NCO operators.
        A unique filename is generated to avoid multithreading errors.
        To overwrite the input file, the source file is dump using the ``cat`` Shell command-line
        to avoid Python memory limit.

        :param str attribute: The attribute to delete
        :param str variable: The variable that has the attribute
        :raises Error: If the deletion failed

        """
        # Generate unique filename
        fftmp = '{0}/{1}{2}'.format(self.directory, str(uuid4()), '.nc')
        try:
            nc = nco.Nco()
            nc.ncatted(input=self.ffp,
                       output=fftmp,
                       options='-a {0},{1},d,,'.format(attribute, variable))
            os.popen("cat {0} > {1}".format(fftmp, self.ffp), 'r')
            os.remove(fftmp)
        except:
            os.remove(fftmp)
            raise NetCDFAttributeRemoveFail(attribute, self.ffp, variable)

    def build_time_axis(self, start, inc, input_units, output_units, calendar, is_instant=False):
        """
        Rebuilds time axis from date axis, depending on MIP frequency, calendar and instant status.

        :param float start: The numerical date to start (from ``netCDF4.date2num`` or \
        :func:`nctime.utils.time.date2num`)
        :param int inc: The time incrementation
        :param input_units: The time units deduced from the frequency
        :param output_units: The time units from the file
        :param calendar: The time calendar fro NetCDF attributes
        :param boolean is_instant: True if instantaneous time axis
        :returns: The corresponding theoretical time axis
        :rtype: *numpy.array*

        """
        num_axis = np.arange(start=start, stop=start + self.length * inc, step=inc)
        if input_units.split(' ')[0] in ['years', 'months']:
            last_date = time.num2date(num_axis[-1], units=input_units, calendar=calendar)[0]
        else:
            last_date = time.num2date(num_axis[-1], units=input_units, calendar=calendar)
        self.last_date = time.date2str(last_date)
        if not is_instant:
            num_axis += 0.5 * inc
        date_axis = time.num2date(num_axis, units=input_units, calendar=calendar)
        return time.date2num(date_axis, units=output_units, calendar=calendar)

    def build_time_bounds(self, start, inc, input_units, output_units, calendar):
        """
        Rebuilds time boundaries from the start date, depending on MIP frequency, calendar and
        instant status.

        :param float start: The numerical date to start (from ``netCDF4.date2num`` or \
        :func:`nctime.utils.time.date2num`)
        :param int inc: The time incrementation
        :param input_units: The time units deduced from the frequency
        :param output_units: The time units from the file
        :param calendar: The time calendar fro NetCDF attributes

        :returns: The corresponding theoretical time boundaries as a [n, 2] array
        :rtype: *numpy.array*

        """
        num_axis_bnds = np.column_stack(((np.arange(start=start,
                                                    stop=start + self.length * inc,
                                                    step=inc)),
                                         (np.arange(start=start,
                                                    stop=start + (self.length + 1) * inc,
                                                    step=inc)[1:])))
        date_axis_bnds = time.num2date(num_axis_bnds, units=input_units, calendar=calendar)
        return time.date2num(date_axis_bnds, units=output_units, calendar=calendar)

    def nc_var_overwrite(self, variable, data):
        """
        Rewrite variable to NetCDF file without copy.

        :param str variable: The variable to replace
        :param float array data: The data array to overwrite

        """
        try:
            f = netCDF4.Dataset(self.ffp, 'r+')
        except IOError:
            raise InvalidNetCDFFile(self.ffp)
        f.variables[variable][:] = data
        f.close()

    def nc_att_overwrite(self, attribute, variable, data):
        """
        Rewrite attribute to NetCDF file without copy.


        :param str attribute: THe attribute to replace
        :param str variable: The variable that has the attribute
        :param str data: The string to add to overwrite

        """
        try:
            f = netCDF4.Dataset(self.ffp, 'r+')
        except IOError:
            raise InvalidNetCDFFile(self.ffp)
        setattr(f.variables[variable], attribute, data)
        f.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Methods used to deal with NetCDF file time axis.

"""

import logging
import re

import netCDF4
import numpy as np
from netcdftime import datetime
from netcdftime import utime

from constants import *
from exceptions import *


class TimeInit(object):
    """
    Encapsulates the time properties from first file into processing context.
    These properties has to be used as reference for all file into the directory.

     * The calendar, the frequency and the realm are read from NetCDF global attributes and \
     use to detect instantaneous time axis,
     * The NetCDF time units attribute has to be unchanged in respect with CF convention and \
     archives designs.

    +-------------------+-------------+---------------------------------+
    | Attribute         | Type        | Description                     |
    +===================+=============+=================================+
    | *self*.calendar   | *str*       | NetCDF calendar attribute       |
    +-------------------+-------------+---------------------------------+
    | *self*.frequency  | *str*       | NetCDF frequency attribute      |
    +-------------------+-------------+---------------------------------+
    | *self*.realm      | *str*       | NetCDF modeling realm attribute |
    +-------------------+-------------+---------------------------------+
    | *self*.tunits     | *str*       | Time units from file            |
    +-------------------+-------------+---------------------------------+
    | *self*.funits     | *str*       | Time units from frequency       |
    +-------------------+-------------+---------------------------------+
    | *self*.is_instant | *boolean*   | True if instantaneous axis      |
    +-------------------+-------------+---------------------------------+
    | *self*.has_bounds | *boolean*   | True if require time bounds     |
    +-------------------+-------------+---------------------------------+

    :param ProcessingContext ctx: The processing context
    :raises Error: If NetCDF time units attribute is missing
    :raises Error: If NetCDF frequency attribute is missing
    :raises Error: If NetCDF realm attribute is missing
    :raises Error: If NetCDF calendar attribute is missing

    """

    def __init__(self, ctx):
        ref_path = '{directory}/{ref}'.format(**ctx.__dict__)
        try:
            nc = netCDF4.Dataset(ref_path, 'r')
        except IOError:
            raise InvalidNetCDFFile(ref_path)
        # Check required global attributes exist
        for attribute in REQUIRED_ATTRIBUTES:
            if attribute not in nc.ncattrs():
                raise NoNetCDFAttribute(attribute, ref_path)
        # Check time variable exists
        if 'time' not in nc.variables.keys():
            raise NoNetCDFVariable('time', ref_path)
        # Check required time attributes exist
        for attribute in REQUIRED_TIME_ATTRIBUTES:
            if attribute not in nc.variables['time'].ncattrs():
                raise NoNetCDFAttribute(attribute, ref_path, 'time')
        # Get realm
        if nc.project_id == 'CORDEX':
            self.realm = 'atmos'
        else:
            self.realm = nc.modeling_realm
        # Get frequency
        self.frequency = nc.frequency
        # Get time units (i.e., days since ...)
        self.tunits = self.control_time_units(nc.variables['time'].units, ctx.tunits_default)
        # Convert time units into frequency units (i.e., months/year/hours since ...)
        self.funits = self.convert_time_units(self.tunits, self.frequency)
        # Get calendar
        self.calendar = nc.variables['time'].calendar
        if self.calendar == 'standard' and nc.model_id == 'CMCC-CM':
            self.calendar = 'proleptic_gregorian'
        # Get boolean on instantaneous time axis
        if 'cell_methods' not in nc.variables[ctx.variable].ncattrs():
            raise NoNetCDFAttribute('cell_methods', ref_path, ctx.variable)
        self.is_instant = False
        if 'point' in nc.variables[ctx.variable].cell_methods.lower():
            self.is_instant = True
        # Get boolean on time boundaries
        self.has_bounds = False
        if 'time_bnds' in nc.variables.keys():
            self.has_bounds = True
        nc.close()

    @staticmethod
    def control_time_units(tunits, tunits_default=None):
        """
        Controls the time units format as at least "days since YYYY-MM-DD".
        The time units can be forced within configuration file using the ``time_units_default`` option.

        :param str tunits: The NetCDF time units string from file
        :param tunits_default: The default time units that should be used
        :returns: The appropriate time units string formatted and controlled depending on the project
        :rtype: *str*

        """
        units = tunits.split()
        units[0] = unicode('days')
        if len(units[2].split('-')) == 1:
            units[2] += '-{0}-{1}'.format('01', '01')
        elif len(units[2].split('-')) == 2:
            units[2] += '-{0}'.format('01')
        if tunits_default and ' '.join(units) != tunits_default:
            logging.warning('Invalid time units. Replace "{0}" by "{1}"'.format(' '.join(units), tunits_default))
            return tunits_default
        else:
            return ' '.join(units)

    @staticmethod
    def convert_time_units(tunits, frequency):
        """
        Converts default time units from file into time units using the MIP frequency.
        As en example, for a 3-hourly file, the time units "days since YYYY-MM-DD" becomes
        "hours since YYYY-MM-DD".

        :param str tunits: The NetCDF time units string from file
        :param str frequency: The time frequency
        :returns: The converted time units string
        :rtype: *str*

        """
        units = {'subhr': 'minutes',
                 '3hr':   'hours',
                 '6hr':   'hours',
                 'day':   'days',
                 'mon':   'months',
                 'yr':    'years'}
        return tunits.replace('days', units[frequency])


def untroncated_timestamp(timestamp):
    """
    Returns proper digits for yearly and monthly truncated timestamps.
    The dates from filename are filled with the 0 digit to reach 14 digits.
    Consequently, yearly dates starts at January 1st and monthly dates starts at first day
    of the month.

    :param str timestamp: A date string from a filename
    :returns: The filled timestamp
    :rtype: *str*

    """
    if len(timestamp) == 4:
        # Start year at january 1st
        return (timestamp + '0101').ljust(14, '0')
    elif len(timestamp) == 6:
        # Start month at first day
        return (timestamp + '01').ljust(14, '0')
    else:
        return timestamp.ljust(14, '0')


def num2date(num_axis, units, calendar):
    """
    A wrapper from ``netCDF4.num2date`` able to handle "years since" and "months since" units.
    If time units are not "years since" or "months since", calls usual ``netcdftime.num2date``.

    :param numpy.array num_axis: The numerical time axis following units
    :param str units: The proper time units
    :param str calendar: The NetCDF calendar attribute
    :returns: The corresponding date axis
    :rtype: *array*

    """
    if not units.split(' ')[0] in ['years', 'months']:
        # If units are not 'years' or 'months since', call usual netcdftime.num2date:
        return netCDF4.num2date(num_axis, units=units, calendar=calendar)
    else:
        # Return to time reference with 'days since'
        units_as_days = 'days ' + ' '.join(units.split(' ')[1:])
        # Convert the time reference 'units_as_days' as datetime object
        start_date = netCDF4.num2date(0.0, units=units_as_days, calendar=calendar)
        # Control num_axis to always get an Numpy array (even with a scalar)
        num_axis_mod = np.atleast_1d(np.array(num_axis))
        if units.split(' ')[0] == 'years':
            # If units are 'years since'
            # Define the number of maximum and minimum years to build a date axis covering
            # the whole 'num_axis' period
            max_years = np.floor(np.max(num_axis_mod)) + 1
            min_years = np.ceil(np.min(num_axis_mod)) - 1
            # Create a date axis with one year that spans the entire period by year
            years_axis = np.array([add_year(start_date, years_to_add)
                                   for years_to_add in np.arange(min_years, max_years + 2)])
            # Convert rebuilt years axis as 'number of days since'
            cdftime = utime(units_as_days, calendar=calendar)
            years_axis_as_days = cdftime.date2num(years_axis)
            # Index of each years
            yind = np.vectorize(np.int)(np.floor(num_axis_mod))
            # Rebuilt num_axis as 'days since' adding the number of days since referenced time
            # with an half-increment (num_axis_mod - yind) = 0 or 0.5
            num_axis_mod_days = (years_axis_as_days[yind - int(min_years)] +
                                 (num_axis_mod - yind) *
                                 np.diff(years_axis_as_days)[yind - int(min_years)])
            # Convert result as date axis
            return netCDF4.num2date(num_axis_mod_days, units=units_as_days, calendar=calendar)
        elif units.split(' ')[0] == 'months':
            # If units are 'months since'
            # Define the number of maximum and minimum months to build a date axis covering
            # the whole 'num_axis' period
            max_months = np.floor(np.max(num_axis_mod)) + 1
            min_months = np.ceil(np.min(num_axis_mod)) - 1
            # Create a date axis with one month that spans the entire period by month
            months_axis = np.array([add_month(start_date, months_to_add)
                                    for months_to_add in np.arange(min_months, max_months + 12)])
            # Convert rebuilt months axis as 'number of days since'
            cdftime = utime(units_as_days, calendar=calendar)
            months_axis_as_days = cdftime.date2num(months_axis)
            # Index of each months
            mind = np.vectorize(np.int)(np.floor(num_axis_mod))
            # Rebuilt num_axis as 'days since' adding the number of days since referenced time
            # with an half-increment (num_axis_mod - mind) = 0 or 0.5
            num_axis_mod_days = (months_axis_as_days[mind - int(min_months)] +
                                 (num_axis_mod - mind) *
                                 np.diff(months_axis_as_days)[mind - int(min_months)])
            # Convert result as date axis
            return netCDF4.num2date(num_axis_mod_days, units=units_as_days, calendar=calendar)


def date2num(date_axis, units, calendar):
    """
    A wrapper from ``netCDF4.date2num`` able to handle "years since" and "months since" units.
    If time units are not "years since" or "months since" calls usual ``netcdftime.date2num``.

    :param numpy.array date_axis: The date axis following units
    :param str units: The proper time units
    :param str calendar: The NetCDF calendar attribute
    :returns: The corresponding numerical time axis
    :rtype: *array*

    """
    # date_axis is the date time axis incremented following units (i.e., by years, months, etc).
    if not units.split(' ')[0] in ['years', 'months']:
        # If units are not 'years' or 'months since', call usual netcdftime.date2num:
        return netCDF4.date2num(date_axis, units=units, calendar=calendar)
    else:
        # Return to time reference with 'days since'
        units_as_days = 'days ' + ' '.join(units.split(' ')[1:])
        # Convert date axis as number of days since time reference
        days_axis = netCDF4.date2num(date_axis, units=units_as_days, calendar=calendar)
        # Convert the time reference 'units_as_days' as datetime object
        start_date = netCDF4.num2date(0.0, units=units_as_days, calendar=calendar)
        # Create years axis from input date axis
        years = np.array([date.year for date in np.atleast_1d(np.array(date_axis))])
        if units.split(' ')[0] == 'years':
            # If units are 'years since'
            # Define the number of maximum and minimum years to build a date axis covering
            # the whole 'num_axis' period
            max_years = np.max(years - start_date.year + 1)
            min_years = np.min(years - start_date.year - 1)
            # Create a date axis with one year that spans the entire period by year
            years_axis = np.array([add_year(start_date, yid)
                                   for yid in np.arange(min_years, max_years + 2)])
            # Convert years axis as number of days since time reference
            cdftime = utime(units_as_days, calendar=calendar)
            years_axis_as_days = cdftime.date2num(years_axis)
            # Find closest index for years_axis_as_days in days_axis
            closest_index = np.searchsorted(years_axis_as_days, days_axis)
            # Compute the difference between closest value of year axis and start date, in number of days
            num = days_axis - years_axis_as_days[closest_index]
            # Number of days of the corresponding closest year
            den = np.diff(years_axis_as_days)[closest_index]
            return min_years + closest_index + num / den
        elif units.split(' ')[0] == 'months':
            # If units are 'months since'
            # Define the number of maximum and minimum months to build a date axis covering
            # the whole 'num_axis' period
            max_months = np.max(12 * (years - start_date.year + 1))
            min_months = np.min(12 * (years - start_date.year - 1))
            # Create a date axis with one month that spans the entire period by month
            months_axis = np.array([add_month(start_date, mid)
                                    for mid in np.arange(min_months, max_months)])
            # Convert months axis as number of days since time reference
            cdftime = utime(units_as_days, calendar=calendar)
            months_axis_as_days = cdftime.date2num(months_axis)
            # Find closest index for months_axis_as_days in days_axis
            closest_index = np.searchsorted(months_axis_as_days, days_axis)
            # Compute the difference between closest value of months axis and start date, in number of days
            num = days_axis - months_axis_as_days[closest_index]
            # Number of days of the corresponding closest month
            den = np.diff(months_axis_as_days)[closest_index]
            return min_months + closest_index + num / den


def add_month(date, months_to_add):
    """
    Finds the next month from date.

    :param netcdftime.datetime date: Accepts datetime or phony datetime from ``netCDF4.num2date``.
    :param int months_to_add: The number of months to add to the date
    :returns: The final date
    :rtype: *netcdftime.datetime*

    """
    years_to_add = int((date.month + months_to_add - np.mod(date.month + months_to_add - 1, 12) - 1) / 12)
    new_month = int(np.mod(date.month + months_to_add - 1, 12)) + 1
    new_year = date.year + years_to_add
    date_next = datetime(year=new_year,
                         month=new_month,
                         day=date.day,
                         hour=date.hour,
                         minute=date.minute,
                         second=date.second)

    return date_next


def add_year(date, years_to_add):
    """
    Finds the next year from date.

    :param netcdftime.datetime date: Accepts datetime or phony datetime from ``netCDF4.num2date``.
    :param int years_to_add: The number of years to add to the date
    :returns: The final date
    :rtype: *netcdftime.datetime*

    """
    new_year = date.year + years_to_add
    date_next = datetime(year=new_year,
                         month=date.month,
                         day=date.day,
                         hour=date.hour,
                         minute=date.minute,
                         second=date.second)
    return date_next


def get_start_end_dates_from_filename(filename, pattern, frequency, calendar):
    """
    Returns datetime objects for start and end dates from the filename.
    To rebuild a proper time axis, the dates from filename are expected to set the first
    time boundary and not the middle of the time interval.

    :param str filename: The filename
    :param re Object pattern: The filename pattern as a regex (from `re library \
    <https://docs.python.org/2/library/re.html>`_).
    :param str frequency: The time frequency
    :param str calendar: The NetCDF calendar attribute
    :param boolean is_instant: True if instantaneous time axis
    :returns: ``datetime`` instances for start and end dates from the filename
    :rtype: *netcdftime.datetime*

    """
    dates = []
    date_as_since = None
    for key in ['start_period', 'end_period']:
        date = re.match(pattern, filename).groupdict()[key]
        digits = untroncated_timestamp(date)
        # Convert string digits to %Y-%m-%d %H:%M:%S format
        date_as_since = ''.join([''.join(triple) for triple in
                                 zip(digits[::2], digits[1::2], ['', '-', '-', ' ', ':', ':', ':'])])[:-1]
        # Use num2date to create netCDF4 datetime objects
        if frequency in ['3hr', '6hr']:
            # Fix on filename digits for 3hr and 6hr frequencies. 3hr (6hr) files always start
            # at 000000 end at 2100000 (180000) whether the time axis is instantaneous or not.
            dates.append(num2date(TIME_CORRECTION[frequency][key][digits[-6:]],
                                  units='days since ' + date_as_since,
                                  calendar=calendar))
            # if is_instant:
            #     dates.append(num2date(INSTANT_TIME_CORRECTION[frequency][key][digits[-6:]],
            #                           units='days since ' + date_as_since,
            #                           calendar=calendar))
            # else:
            #     dates.append(num2date(AVERAGED_TIME_CORRECTION[frequency][key][digits[-6:]],
            #                           units='days since ' + date_as_since,
            #                           calendar=calendar))
        else:
            dates.append(num2date(0.0, units='days since ' + date_as_since, calendar=calendar))
    # Append date next to the end date for overlap diagnostic
    try:
        dates.append(num2date(time_inc(frequency)[0],
                              units=time_inc(frequency)[1] + ' since ' + date_as_since,
                              calendar=calendar)[0])
    except TypeError:
        dates.append(num2date(time_inc(frequency)[0],
                              units=time_inc(frequency)[1] + ' since ' + date_as_since,
                              calendar=calendar))
    return dates


def time_inc(frequency):
    """
    Returns the time incrementation and time units depending on the MIP frequency.
    :param str frequency: The MIP frequency
    :returns: The corresponding time value and units
    :rtype: *list*
    """
    inc = {'subhr': [30, 'minutes'],
           '3hr':   [3, 'hours'],
           '6hr':   [6, 'hours'],
           'day':   [1, 'days'],
           'mon':   [1, 'months'],
           'yr':    [1, 'years']}
    return inc[frequency]


def dates2str(dates, sep=True):
    """
    Prints date in format: %Y%m%d %H:%M:%s.

    :param list dates: A list of datetime or phony datetime objects
    :param boolean sep: Print corresponding date with separators if True
    :returns: The corresponding formatted date to print
    :rtype: *str*

    """
    strdates = []
    for date in dates:
        if sep:
            strdates.append('{0:04d}-{1:02d}-{2:02d} {3:02d}:{4:02d}:{5:02d}'.format(date.year,
                                                                                     date.month,
                                                                                     date.day,
                                                                                     date.hour,
                                                                                     date.minute,
                                                                                     date.second))
        else:
            strdates.append('{0:04d}{1:02d}{2:02d}{3:02d}{4:02d}{5:02d}'.format(date.year,
                                                                                date.month,
                                                                                date.day,
                                                                                date.hour,
                                                                                date.minute,
                                                                                date.second))
    return strdates


def date2str(date, sep=True):
    """
    Prints date in format: %Y%m%d %H:%M:%s.

    :param netcdftime.datetime date: A datetime or phony datetime objects
    :param boolean sep: Print date with separators if True
    :returns: The corresponding formatted date to print
    :rtype: *str*

    """
    if sep:
        return '{0:04d}-{1:02d}-{2:02d} {3:02d}:{4:02d}:{5:02d}'.format(date.year,
                                                                        date.month,
                                                                        date.day,
                                                                        date.hour,
                                                                        date.minute,
                                                                        date.second)
    else:
        return '{0:04d}{1:02d}{2:02d}{3:02d}{4:02d}{5:02d}'.format(date.year,
                                                                   date.month,
                                                                   date.day,
                                                                   date.hour,
                                                                   date.minute,
                                                                   date.second)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. module:: nctime
    :platform: Unix
    :synopsis: Toolbox to diagnose NetCDF time axis.

.. moduleauthor:: Guillaume Levavasseur <glipsl@ipsl.jussieu.fr>

"""

import argparse
import os
from datetime import datetime

from utils.utils import MultilineFormatter, init_logging, directory_checker, path_switcher

# Program version
__version__ = 'v{0} {1}'.format('3.9.3', datetime(year=2016, month=12, day=23).strftime("%Y-%d-%m"))


def get_args():
    """
    Returns parsed command-line arguments. See ``nctime -h`` for full description.

    :returns: The corresponding ``argparse`` Namespace
    :rtype: *ArgumentParser*

    """
    #############################
    # Main parser for "esgprep" #
    #############################
    main = argparse.ArgumentParser(
        prog='nctime',
        description="""
        NetCDF files describe all dimensions necessary to work with. In the climate community, this format is widely
        used following the `CF conventions <http://cfconventions.org/>`_. Dimensions such as longitude, latitude and
        time are included in NetCDF files as vectors.|n|n

        The time axis is a key dimension. Unfortunately, this time axis often is mistaken in files from coupled climate
        models and leads to flawed studies or unused data. Consequently, these files cannot be used or, even worse,
        produced erroneous results, due to problems in the time axis description.|n|n

        Moreover, to produce smaller files, the NetCDF files are splitted over the time period. Consequently, the
        different MIP archives designs include the period dates into the filename. The scheme of chunked files in MIP
        projects is not fixed and depends on several parameters (the institute, the model, the frequency, etc.). These
        different schemes lead to unnecessary overlapping files with a more complex folder reading and wasting disk
        space.|n|n

        "nctime" is a Python toolbox allowing you to easily:|n
         - Highlight chunked NetCDF files producing overlaps in a time series and delete all chunked overlapping
         files,|n
         - Check and rebuild a MIP-compliant time axis of your NetCDF files.|n|n

        Note that "nctime" is based on uncorrupted filename period dates and properly-defined times units, time
        calendar and frequency NetCDF attributes.|n|n

        See full documentation and references on http://nctime.readthedocs.io/.
        """,
        formatter_class=MultilineFormatter,
        add_help=False,
        epilog="""
        Developed by:|n
        Levavasseur, G. (UPMC/IPSL - glipsl@ipsl.jussieu.fr)
        """)
    main._optionals.title = "Optional arguments"
    main._positionals.title = "Positional arguments"
    main.add_argument(
        '-h', '--help',
        action='help',
        help="""Show this help message and exit.""")
    main.add_argument(
        '-V',
        action='version',
        version='%(prog)s ({0})'.format(__version__),
        help="""Program version.""")
    subparsers = main.add_subparsers(
        title='Tools as subcommands',
        dest='cmd',
        metavar='',
        help='')

    #######################################
    # Parent parser with common arguments #
    #######################################
    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument(
        'directory',
        type=directory_checker,
        nargs='?',
        help="""Variable path to diagnose.""")
    parent.add_argument(
        '--project',
        metavar='<project_id>',
        type=str,
        required=True,
        help="""
        Required project name corresponding to a section of the|n
        configuration file.
        """)
    parent.add_argument(
        '-i',
        metavar='$PWD/config.ini',
        type=str,
        default='{0}/config.ini'.format(os.getcwd()),
        help="""Path of configuration INI file.""")
    parent.add_argument(
        '--log',
        metavar='<log_dir>',
        type=str,
        const='NO_PATH',
        nargs='?',
        help="""
        Logfile directory. <log_dir> is read from configuration |n
        file if not submitted. If no <log_dir>, current working |n
        directory is used instead. An existing logfile can be submitted.|n
        If not, standard output is used.
        """)
    parent.add_argument(
        '-h', '--help',
        action='help',
        help="""Show this help message and exit.""")
    parent.add_argument(
        '-v',
        action='store_true',
        default=False,
        help="""Verbose mode.""")

    ##################################
    # Subparser for "nctime overlap" #
    ##################################
    overlap = subparsers.add_parser(
        'overlap',
        prog='nctime overlap',
        description="""
        To produce smaller files, the NetCDF files are splited over the time period. Consequently, the different MIP
        archives designs include the period dates into the filename.|n|n

        The scheme of chunked files in MIP projects is not fixed and depends on several parameters (the institute,
        the model, the frequency, etc.). These different schemes lead to unnecessary overlapping files with a more
        complex folder reading and wasting disk space.|n|n

        "overlap" is a command-line tool allowing you to easily highlight chunked NetCDF files producing overlaps in
        a time series and delete all chunked overlapping files in your MIP variable directories in order to save
        disk space.|n|n

        Note that:|n (i) Only complete overlaps are detected. For example, if a file goes from 1991 to 2010 and
        another goes from 2001 to 2020, the overap is partial. If the second file goes from 2001 to 2010 so the
        overlap is complete and the second file can be removed without loss of information.|n (ii) "overlap" is
        based on uncorrupted filename period dates and properly-defined times units, time calendar and frequency
        NetCDF attributes.|n|n

        The default values are displayed next to the corresponding flags.
        """,
        formatter_class=MultilineFormatter,
        help="""
        Highlight chunked NetCDF files producing overlap in a time series.|n
        See "nctime overlap -h" for full help.""",
        add_help=False,
        parents=[parent])
    overlap._optionals.title = "Optional arguments"
    overlap._positionals.title = "Positional arguments"
    overlap.add_argument(
        '--mip',
        metavar='*',
        nargs='?',
        default='*',
        help="""
        Specifies a MIP table value so as to filter directory|n
        content (Unix wildcard are allowed).
        """)
    overlap.add_argument(
        '--remove',
        action='store_true',
        default=False,
        help="""
        Removes overlapping files.|n
        THIS ACTION DEFINITELY MODIFY INPUT DIRECTORY!
        """)
    overlap.add_argument(
        '--subtree',
        action='store_true',
        default=False,
        help="""If a period gap, use the sub-period from the start date.""")

    ###############################
    # Subparser for "nctime axis" #
    ###############################
    axis = subparsers.add_parser(
        'axis',
        prog='nctime axis',
        description="""
        NetCDF files describe all dimensions necessary to work with. In the climate community, this format is widely
        used following the CF conventions. Dimensions such as longitude, latitude and time are included in NetCDF
        files as vectors.|n|n

        The time axis is a key dimension. Unfortunately, this time axis often is mistaken in files from coupled
        climate models and leads to flawed studies or unused data.|n|n

        "time_axis" is a command-line tool allowing you to easily check and rebuild a MIP-compliant time axis of
        your downloaded files from the ESGF.|n|n

        Note that:|n
        (i) "time_axis" is based on uncorrupted filename period dates and properly-defined times units, time
        calendar and frequency NetCDF attributes.|n
        (ii) To rebuild a proper time axis, the dates from filename are expected to set the first time boundary and
        not the middle of the time interval. This is always the case for the instantaneous axis or frequencies
        greater than the daily frequency. Consequently, the 3-6 hourly files with an averaged time axis requires a
        date time correction.|n|n

        Time axis status returned:|n
        000: Unmodified time axis,|n
        001: Corrected time axis because wrong time steps,|n
        002: Corrected time axis because of changing time units,|n
        003: Ignored time axis because of inconsistency between last date of time axis and end date of filename
        period (e.g., wrong time axis length),|n
        004: Corrected time axis deleting time boundaries for instant time,|n
        005: Ignored averaged time axis without time boundaries,|n
        006: Corrected time bounds because wrong time steps.|n|n

        The default values are displayed next to the corresponding flags.
        """,
        formatter_class=MultilineFormatter,
        help="""
        Rewrite and/or check time axis of MIP NetCDF files.|n
        See "nctime axis -h" for full help.
        """,
        add_help=False,
        parents=[parent])
    axis._optionals.title = "Optional arguments"
    axis._positionals.title = "Positional arguments"
    axis.add_argument(
        '--write',
        action='store_true',
        default=False,
        help="""
        Rewrites time axis depending on checking.|n
        THIS ACTION DEFINITELY MODIFY INPUT FILES!
        """)
    axis.add_argument(
        '--force',
        action='store_true',
        default=False,
        help="""
        Forces time axis writing overpassing checking step.|n
        THIS ACTION DEFINITELY MODIFY INPUT FILES!
        """)
    axis.add_argument(
        '--db',
        metavar='<db_path>',
        type=str,
        nargs='?',
        const='NO_PATH',
        help="""
        SQLite database path to persist diagnostics. <db_path> is |n
        read from configuration file if not submitted. If no <db_path>, |n
        current working directory is used instead. If not, time |n
        diagnostic is not saved.
        """)
    axis.add_argument(
        '--max-threads',
        metavar='<max_threads>',
        type=int,
        default=0,
        help="""
        Number of maximal threads to simultaneously process several |n
        files. If not, <max_threads> from configuration file is used.|n
        If no default <max_threads>, only one thread is used (i.e., |n
        sequential processing).
        """)

    return main.parse_args()


def run():
    # Get command-line arguments
    args = get_args()
    # Initialize logger
    args.log = path_switcher('log', args)
    if args.v:
        init_logging(args.log, level='DEBUG')
    else:
        init_logging(args.log)

    # Run subcommand
    if args.cmd == 'axis':
        from axis import main
        main.main(args)
    elif args.cmd == 'overlap':
        from overlap import main
        main.main(args)


# Main entry point for stand-alone call.
if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
   :platform: Unix
   :synopsis: Useful functions to use with this package.

"""

import ConfigParser
import logging
import os
import re
import textwrap
from argparse import HelpFormatter, ArgumentTypeError
from datetime import datetime

from constants import *
from exceptions import *


class MultilineFormatter(HelpFormatter):
    """
    Custom formatter class for argument parser to use with the Python
    `argparse <https://docs.python.org/2/library/argparse.html>`_ module.

    """

    def __init__(self, prog):
        # Overload the HelpFormatter class.
        super(MultilineFormatter, self).__init__(prog, max_help_position=60, width=100)

    def _fill_text(self, text, width, indent):
        # Rewrites the _fill_text method to support multiline description.
        text = self._whitespace_matcher.sub(' ', text).strip()
        multiline_text = ''
        paragraphs = text.split('|n|n ')
        for paragraph in paragraphs:
            lines = paragraph.split('|n ')
            for line in lines:
                formatted_line = textwrap.fill(line, width,
                                               initial_indent=indent,
                                               subsequent_indent=indent) + '\n'
                multiline_text += formatted_line
            multiline_text += '\n'
        return multiline_text

    def _split_lines(self, text, width):
        # Rewrites the _split_lines method to support multiline helps.
        text = self._whitespace_matcher.sub(' ', text).strip()
        lines = text.split('|n ')
        multiline_text = []
        for line in lines:
            multiline_text.append(textwrap.fill(line, width))
        multiline_text[-1] += '\n'
        return multiline_text


class CfgParser(ConfigParser.ConfigParser):
    """
    Custom ConfigParser class

    """

    def __init__(self):
        ConfigParser.ConfigParser.__init__(self)
        self.read_paths = list()

    def options(self, section, defaults=True, **kwargs):
        """
        Can get options() with/without defaults.

        """
        try:
            opts = self._sections[section].copy()
        except KeyError:
            raise ConfigParser.NoSectionError(section)
        if defaults:
            opts.update(self._defaults)
        if '__name__' in opts:
            del opts['__name__']
        return opts.keys()

    def read(self, filenames):
        """
        Read and parse a filename or a list of filenames, and records there paths.
        """
        if isinstance(filenames, basestring):
            filenames = [filenames]
        for filename in filenames:
            try:
                fp = open(filename)
            except IOError:
                continue
            self._read(fp, filename)
            fp.close()
            self.read_paths.append(filename)


def init_logging(log, level='INFO'):
    """
    Initiates the logging configuration (output, date/message formatting).
    If a directory is submitted the logfile name is unique and formatted as follows:
    ``name-YYYYMMDD-HHMMSS-JOBID.log``If ``None`` the standard output is used.

    :param str log: The logfile name or directory.
    :param str level: The log level.

    """
    log_fmt = '%(asctime)s %(levelname)s %(message)s'
    log_date_fmt = '%Y/%m/%d %I:%M:%S %p'
    log_levels = {'CRITICAL': logging.CRITICAL,
                      'ERROR':    logging.ERROR,
                      'WARNING':  logging.WARNING,
                      'INFO':     logging.INFO,
                      'DEBUG':    logging.DEBUG,
                      'NOTSET':   logging.NOTSET}
    if log:
        if os.path.isfile(log):
            logging.basicConfig(filename=log,
                                level=log_levels[level],
                                format=log_fmt,
                                datefmt=log_date_fmt)
        else:
            logfile = 'timeaxis-{0}-{1}.log'.format(datetime.now().strftime("%Y%m%d-%H%M%S"),
                                                   os.getpid())
            if not os.path.isdir(log):
                os.makedirs(log)
            logging.basicConfig(filename=os.path.join(log, logfile),
                                level=log_levels[level],
                                format=log_fmt,
                                datefmt=log_date_fmt)
    else:
        logging.basicConfig(level=log_levels[level],
                            format=log_fmt,
                            datefmt=log_date_fmt)


def directory_checker(path):
    """
    Checks if the supplied directory exists. The path is normalized without trailing slash.

    :param str path: The path list to check
    :returns: The same path list
    :rtype: *str*
    :raises Error: If one of the directory does not exist

    """
    path = os.path.abspath(os.path.normpath(path))
    if not os.path.isdir(path):
        msg = "No such directory: '{0}'".format(path)
        raise ArgumentTypeError(msg)
    return path


def path_switcher(flag, args, default=os.getcwd()):
    """
    Switch to the appropriate path depending on the flag value.
    If simple flag without path, the path is read from configuration file.
    If no path is declared into configuration file, a default path is set.
    If the flag is combined with a path, this path is used.
    No flag is no path.

    :param str flag: The argument to evaluate
    :param *ArgumentParser* args: The corresponding ``argparse`` Namespace
    :param str default: The default path to set if not supplied
    :returns: The appropriate path
    :rtype: *str*
    """
    cfg = config_parse(args.i, args.project)
    option = '{0}_path'.format(flag)
    if args.__dict__[flag] == 'NO_PATH':
        if cfg.has_option('DEFAULT', option) and cfg.get('DEFAULT', option) != '':
            return cfg.get('DEFAULT', option)
        else:
            return default
    else:
        return args.__dict__[flag]


def config_parse(path, section):
    """
    Parses the configuration file if exists. Tests if required options are declared.

    :param str path: The absolute or relative path of the configuration file
    :param str section: The section to parse
    :returns: The configuration file parser
    :rtype: *ConfigParser*
    :raises Error: If no configuration file exists or is empty
    :raises Error: If section is missing
    :raises Error: If required options are missing
    :raises Error: If the configuration file parsing fails

    """
    path = os.path.abspath(os.path.normpath(path))
    if not os.path.isfile(path):
        raise NoConfigFile(path)
    cfg = CfgParser()
    cfg.read(path)
    if section not in cfg.sections():
        raise NoConfigSection(section, path)
    if not cfg:
        raise EmptyConfigFile(path)
    for option in REQUIRED_OPTIONS:
        if not cfg.has_option(section, option):
            raise NoConfigOption(option, section, cfg.read_paths)
    return cfg


def translate_filename_format(cfg, project_section):
    """
    Return a list of regular expression filters associated with the ``directory_format`` option
    in the configuration file. This can be passed to the Python ``re`` methods.

    :param RawConfigParser cfg: The configuration file parser (as a :func:`ConfigParser.RawConfigParser` class instance)
    :param str project_section: The project section name to parse
    :returns: The corresponding ``re`` pattern

    """
    # Start translation
    pattern = cfg.get(project_section, 'filename_format', raw=True).strip()
    pattern = pattern.replace('\.', '__ESCAPE_DOT__')
    pattern = pattern.replace('.', r'\.')
    pattern = pattern.replace('__ESCAPE_DOT__', r'\.')
    # Translate period dates
    pattern = re.sub(re.compile(r'%\((start_period)\)s'), r'(?P<\1>[\d]+)', pattern)
    pattern = re.sub(re.compile(r'%\((end_period)\)s'), r'(?P<\1>[\d]+)', pattern)
    # Translate all patterns matching %(name)s
    pattern = re.sub(re.compile(r'%\(([^()]*)\)s'), r'(?P<\1>[\w.-]+)', pattern)
    return '^{0}$'.format(pattern)


def trunc(f, n):
    """
    Truncates a float f to n decimal places before rounding

    :param float f: The number to truncates
    :param int n: Decimal number to place before rounding
    :returns: The corresponding truncated number
    :rtype: *float*

    """
    slen = len('%.*f' % (n, f))
    return float(str(f)[:slen])


def find_key(dic, val):
    """
    Returns the dictionary key given the value.

    :param str val: The value to search
    :param dict dic: The dictionnary
    :returns: The dictionnary key corresponding to the value
    :rtype: *str*

    """
    return [k for k, v in dic.iteritems() if v == val][0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Custom exceptions used in this package.

"""


class EmptyConfigFile(Exception):
    """
    Raised when configuration file is empty.

    """

    def __init__(self, paths):
        self.msg = "Empty configuration parser."
        for path in paths:
            self.msg += "\n<config file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NoConfigFile(Exception):
    """
    Raised when no configuration file found.

    """

    def __init__(self, path):
        self.msg = "No or not a file"
        self.msg += "\n<config file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NoConfigSection(Exception):
    """
    Raised when no corresponding section found in configuration file.

    """

    def __init__(self, section, paths):
        self.msg = "No section: '{0}'".format(section)
        for path in paths:
            self.msg += "\n<config file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NoConfigOption(Exception):
    """
    Raised when no corresponding option found in section of the configuration file.

    """

    def __init__(self, option, section, paths):
        self.msg = "No option: '{0}'".format(option)
        self.msg += "\n<section: '{0}'>".format(section)
        for path in paths:
            self.msg += "\n<config file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NoConfigValue(Exception):
    """
    Raised when no corresponding value found in option of the section from the configuration file.

    """

    def __init__(self, value, option, section, paths):
        self.msg = "No value: '{0}'".format(value)
        self.msg += "\n<option: '{0}'>".format(option)
        self.msg += "\n<section: '{0}'>".format(section)
        for path in paths:
            self.msg += "\n<config file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)


class InvalidNetCDFFile(Exception):
    """
    Raised when not a NetCDF file.

    """

    def __init__(self, path):
        self.msg = "Invalid or unknown file format. Only support NetCDF file."
        self.msg += "\n<file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NoNetCDFAttribute(Exception):
    """
    Raised when a NetCDF attribute is missing.

    """

    def __init__(self, attribute, path, variable=None):
        if variable:
            self.msg = "No attribute: '{0}'".format(attribute)
            self.msg += "\n<variable: '{0}'>".format(variable)
        else:
            self.msg = "No global attribute: '{0}'".format(attribute)
        self.msg += "\n<file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NoNetCDFVariable(Exception):
    """
    Raised when a NetCDF variable is missing.

    """

    def __init__(self, variable, path):
        self.msg = "No variable: '{0}'".format(variable)
        self.msg += "\n<file: '{0}'>".format(path)
        super(self.__class__, self).__init__(self.msg)
